from google.cloud import bigquery


class BQ(object):
    def __init__(self, dataset_id=None, table_id=None, project_id=None, cred_file_path=None):
        self.dataset_id = dataset_id
        self.table_id = table_id
        if cred_file_path is None:
            self.client = bigquery.Client(project=project_id)
        else:
            self.client = bigquery.Client.from_service_account_json(cred_file_path)
        if project_id is not None:
            self.client.project = project_id

    def __dataset_id(self, dataset_id):
        if dataset_id is None:
            dataset_id = self.dataset_id
        return dataset_id

    def __table_id(self, table_id):
        if table_id is None:
            table_id = self.table_id
        return table_id

    def create_dataset(self, dataset_id=None, location='EU'):
        dataset_id = self.__dataset_id(dataset_id)
        dataset_ref = self.client.dataset(dataset_id)
        dataset = bigquery.Dataset(dataset_ref)
        dataset.location = location
        self.client.create_dataset(dataset)  # API request

    def create_table(self, schema, dataset_id=None, table_id=None, partitioning_field=None, clustering_fields=[], require_partition_filter=False):
        dataset_id = self.__dataset_id(dataset_id)
        table_id = self.__table_id(table_id)

        if self.check_if_table_exists(table_id) is False:
            dataset_ref = self.client.dataset(dataset_id)
            table_ref = dataset_ref.table(table_id)
            table = bigquery.Table(table_ref, schema=schema)
            if partitioning_field is not None:
                if partitioning_field == '_PARTITIONTIME':
                    table.time_partitioning = bigquery.table.TimePartitioning(type_=bigquery.TimePartitioningType.DAY, require_partition_filter=require_partition_filter, field=None)
                else:
                    table.time_partitioning = bigquery.table.TimePartitioning(type_=bigquery.TimePartitioningType.DAY, require_partition_filter=require_partition_filter, field=partitioning_field)
            if len(clustering_fields) > 0:
                table.clustering_fields = clustering_fields
            table = self.client.create_table(table)  # API request
            _ = table.table_id
        else:
            print("Table {} already exists or another error occurred.".format(table_id))

    def check_if_table_exists(self, dataset_id=None, table_id=None):
        dataset_id = self.__dataset_id(dataset_id)
        table_id = self.__table_id(table_id)
        dataset_ref = self.client.dataset(dataset_id)
        dataset = bigquery.Dataset(dataset_ref)
        table_ref = dataset.table(table_id)
        try:
            table = self.client.get_table(table_ref)  # API request
            _ = table.schema
            # If no error, table exists
            table_exists = True
        except:
            table_exists = False
        return table_exists

    def get_schema(self, dataset_id=None, table_id=None):
        dataset_id = self.__dataset_id(dataset_id)
        table_id = self.__table_id(table_id)
        dataset_ref = self.client.dataset(dataset_id)
        dataset = bigquery.Dataset(dataset_ref)
        table_ref = dataset.table(table_id)
        table = self.client.get_table(table_ref) # API request
        return table.schema

    def streaming_insert_single_json(self, data, dataset_id=None, table_id=None):
        dataset_id = self.__dataset_id(dataset_id)
        table_id = self.__table_id(table_id)
        table_ref = self.client.dataset(dataset_id).table(table_id)
        table = self.client.get_table(table_ref)  # API request
        errors = self.client.insert_rows_json(table, [data])  # API request
        return errors

    def run_query(self, query):
        query_job = self.client.query(query)  # API request
        rows = query_job.result()  # Waits for query to finish
        return rows.to_dataframe()



if __name__ == '__main__':

    # With default values: (i. e. if you only work within one dataset)
    bq_handler = BQ(dataset_id='default_dataset_id', table_id='default_table_id')
    # Without defaults:
    # bq_handler = BQ()

    # You can also specify a project_id and a cred_file_path (credential file path) if you need it

    # Creating a new Dataset:
    #bq_handler.create_dataset(dataset_id='test_dataset_id')

    # Creating a new Table:
    from google.cloud import bigquery
    SCHEMA = [
        bigquery.SchemaField('full_name', 'STRING',
                             mode='required', description="Visitor's Name"),
        bigquery.SchemaField('visit_time', 'TIMESTAMP',
                             mode='required', description="Visit Time"),
        bigquery.SchemaField('visit_length', 'INT64',
                             mode='required', description="Length of Visit in Seconds"),
        bigquery.SchemaField('sentiment', 'FLOAT64',
                             mode='required', description="Calculated Happiness Score"),
    ]
    #bq_handler.create_table(schema=SCHEMA, dataset_id='test_dataset_id', table_id='test_table_id')
    # You can also create a table with day partitioning and with clustering:
    #bq_handler.create_table(schema=SCHEMA, dataset_id='test_dataset_id', table_id='test_table_id', partitioning_field='_PARTITIONTIME', require_partition_filter=False, clustering_fields=['visit_time'])
    # _PARTITIONTIME is the default partitioning from BigQuery


    # Checking if a table exists or not:
    #check = bq_handler.check_if_table_exists(dataset_id='test_dataset_id', table_id='test_table_id')
    #print(check)

    # Getting a schema of an existing table:
    #schema = bq_handler.get_schema(dataset_id='test_dataset_id', table_id='test_table_id')
    #print(schema)

    # Streaming a row into an existing BQ table:
    data = {'full_name': 'Max Mustermann', 'visit_time': '2019-07-23 13:45:07.372826 UTC', 'visit_length': 100, 'sentiment': 1.111}
    bq_handler.streaming_insert_single_json(data, dataset_id='test_dataset_id', table_id='test_table_id')

    # Running a query and get results as a dataframe:
    query = '''
    SELECT * FROM `v135-5683-playground-goppold.test_dataset_id.test_table_id`  
    '''
    df = bq_handler.run_query(query)
    print(df)




from mms.logger.cloud_run_logger import CloudRunLogger


if __name__ == '__main__':

    logger = CloudRunLogger(service_name='th-testservice-new',
                            trace_id='1234',
                            project_id='spielwiese-tobias',
                            revision_version='latese',
                            location='europe-west1',
                            local_run=True)


    logger.error('This is a test error')
    logger.info('This is a test info')




import json
import sys


class CloudRunLogger(object):

    def __init__(self, project_id, local_run=False):
        self.project_id = project_id
        self.request = None
        self.trace_id = None
        self.local_run = local_run

    def update_trace_id(self, trace_id):
        self.trace_id = trace_id

    def update_request_header(self, api_request):
        self.request = api_request

    def _do_log(self, severity, message):
        # check if local run = true if yes just print the message
        if self.local_run:
            print(message)
        else:
            # Build structured log messages as an object.
            global_log_fields = {}
            # Add log correlation to nest all log messages
            # beneath request log in Log Viewer.
            trace_header = self.request.headers.get('X-Cloud-Trace-Context')

            if trace_header and self.project_id:
                trace = trace_header.split('/')
                global_log_fields['logging.googleapis.com/trace'] = (
                    f"projects/{self.project_id}/traces/{trace[0]}")

            # Complete a structured log entry.
            entry = dict(severity=severity,
                         message=message,
                         trace_id=self.trace_id,
                         **global_log_fields)

            print(json.dumps(entry))
            sys.stdout.flush()

    def info(self, message):
        self._do_log('INFO', message='MMS-INFO: {}'.format(message))

    def warning(self, message):
        self._do_log('WARNING', message='MMS-WARNING: {}'.format(message))

    def error(self, message):
        self._do_log('ERROR', message='MMS-ERROR: {}'.format(message))

    def debug(self, message):
        self._do_log('DEBUG', message='MMS-DEBUG: {}'.format(message))




















# Imports the Google Cloud client library
from google.cloud import datastore


class Datastore(object):
    def __init__(self, project_id=None):
        self.client = datastore.Client(project=project_id)

    def put_new_entity(self, kind, prop_df, name=None):
        # The Cloud Datastore key for the new entity
        if name is None:
            task_key = self.client.key(kind)
        else:
            task_key = self.client.key(kind, name)
        entity = datastore.Entity(key=task_key)
        for i in prop_df:
            entity[i] = prop_df[i]
        self.client.put(entity)

    def delete_entity(self, kind, name):
        key = self.client.key(kind, name)
        self.client.delete(key)

    def get_entity(self, kind, name):
        key = self.client.key(kind, name)
        return self.client.get(key=key)

    def query(self, kind, filter):
        query = self.client.query(kind=kind)
        if isinstance(filter[0], list) is False:
            filter = [filter]
        for item in filter:
            query.add_filter(item[0], item[1], item[2])
        result = query.fetch()
        res = result.pages
        results = []
        for i in list(res):
            for j in i:
                results.append(j)
        return results

    def update_entity(self, kind, name, prop_df):
        self.put_new_entity(kind, prop_df, name)

    def get_all_of_a_kind(self, kind):
        query = self.client.query(kind=kind)
        result = query.fetch()
        res = result.pages
        results = []
        for i in list(res):
            for j in i:
                results.append(j)
        return results





if __name__ == '__main__':

    # Examples:
    ds_handler = Datastore()
    ''' 
    # Create new entity on Datastore:
    
    kind = "TestKind1"
    data = {
        "property1": 1,
        "property2": "Property Number 2",
        "property3": 3.33
    }
    name = "TestName or ID"  # if None -> autogenerated from Google
    # ds_handler.put_new_entity(kind=kind, prop_df=data, name=name)

    # Delete Entity:
    # ds_handler.delete_entity(kind="TestKind1", name="TestName or ID")

    # Get Entity from Datastore:
    # entity = ds_handler.get_entity(kind="TestKind1", name="TestName or ID")
    # print(entity)

    # Query datastore:
    with_one_filter = ["property1", "=", 1]
    with_more_filter = [with_one_filter, ["property3", "<=", 3.33]]
    result = ds_handler.query(kind="TestKind1", filter=with_more_filter)
    print(result)


    # Update Entity:
    data = {
        "property1": 0,
        "property2": "Property Number 2",
        "property3": 333,
        "property4": 4
    }
    # ds_handler.update_entity(kind="TestKind1", name="TestName", prop_df=data)
    '''


import time
import jwt
import json
import requests
import urllib

# https://medium.com/@stephen.darling/oauth2-authentication-with-google-cloud-run-700015a092c2


# TODO maybe see
#  - https://medium.com/google-cloud/authenticating-using-google-openid-connect-tokens-e7675051213b
#  - https://cloud.google.com/iam/docs/reference/credentials/rest/v1/projects.serviceAccounts/generateAccessToken


class APICalls(object):

    def __init__(self, service_url, sa_key_dict=None, sa_key_path=None):
        self.service_url = service_url
        self.sa_key_dict = sa_key_dict
        self.sa_key_path = sa_key_path

        if sa_key_dict is None and sa_key_path is None:
            raise ValueError("You have to specify either sa_key_dict or sa_key_path")

        if sa_key_dict is None:
            self.credentials = self.__read_sa_file()
        else:
            self.credentials = sa_key_dict

        self.__exp = None
        self.__signed_jwt = None
        self.__id_token = None

        self.__update_token()

    def __read_sa_file(self):
        with open(self.sa_key_path) as json_file:
            data = json.load(json_file)
        return data

    # create access token we can use as a Bearer token in future REST requests:
    def __exchange_jwt_for_token(self):
        body = {
            'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
            'assertion': self.__signed_jwt
        }
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        token_request = requests.post(
            url='https://www.googleapis.com/oauth2/v4/token',
            headers=headers,
            data=urllib.parse.urlencode(body)
        )
        result = token_request.json()
        self.__id_token = result['id_token']

    # create the locally-signed JWT:
    def __create_signed_jwt(self):
        iat = time.time()
        exp = iat + 3600
        payload = {
            'iss': self.credentials['client_email'],
            'sub': self.credentials['client_email'],
            'target_audience': self.service_url,
            'aud': 'https://www.googleapis.com/oauth2/v4/token',
            'iat': iat,
            'exp': exp
        }
        additional_headers = {'kid': self.credentials['private_key_id']}
        self.__signed_jwt = jwt.encode(payload, self.credentials['private_key'], headers=additional_headers, algorithm='RS256')

    def __get_token_info(self):
        url = "https://oauth2.googleapis.com/tokeninfo?id_token={}".format(self.__id_token)
        r = requests.get(url=url)
        return json.loads(r.content.decode('utf-8'))

    def __update_token(self):
        time_now = time.time()
        if (self.__id_token is None) or (time_now >= self.__exp):
            self.__create_signed_jwt()
            self.__exchange_jwt_for_token()
            token_info = self.__get_token_info()
            self.__exp = float(token_info.get("exp"))

    def get_request(self, url):
        self.__update_token()
        return requests.get(url=url, headers={'Authorization': f'Bearer {self.__id_token}'})

    def post_request(self, url, request_body):
        self.__update_token()
        return requests.post(url=url,  data=json.dumps(request_body), headers={'Authorization': f'Bearer {self.__id_token}'})


if __name__ == '__main__':

    run_service_url = "https://test-api-h3e6iof3xq-ew.a.run.app/"
    auth_handler = APICalls(service_url=run_service_url, sa_key_path="sa-file.json")

    r1 = auth_handler.get_request(url="https://test-api-h3e6iof3xq-ew.a.run.app/apis/test2")

    request_body = {"test_dict:": 1, "test_dict1:": "Dast ist ein Test"}

    r4 = auth_handler.post_request(url="https://test-api-h3e6iof3xq-ew.a.run.app/apis/test1", request_body=request_body)
    response_body = json.loads(r4.content.decode('utf-8'))
    status = r4.status_code


    print("test")


import requests
import google.auth.transport.requests as r
from google.oauth2 import id_token
import time
import pickle


class GCPJWT(object):
    def __init__(self, receiving_service_url):
        self.receiving_service_url = receiving_service_url
        self.metadata_server_token_url = 'http://metadata/computeMetadata/v1/instance/service-accounts/default/identity?audience='

    def _create_jwt_token(self):
        # Set up metadata server request
        # See https://cloud.google.com/compute/docs/instances/verifying-instance-identity#request_signature
        token_request_url = self.metadata_server_token_url + self.receiving_service_url
        token_request_headers = {'Metadata-Flavor': 'Google'}
        # Fetch the token
        token_response = requests.get(token_request_url, headers=token_request_headers)
        jwt = token_response.content.decode("utf-8")
        print(jwt)
        return jwt

    def _verify_token(self, token, get_only_exp=True):
        request = r.Request()
        payload = id_token.verify_token(token, request=request, audience=self.receiving_service_url)
        if get_only_exp is True:
            return payload.get('exp')
        else:
            return payload

    def get_token(self):
        token = self._create_jwt_token()
        exp = self._verify_token(token)
        return token, exp

    def check_if_token_is_exp(self, expiration_unix_time):
        unix_time_now = int(time.time())
        if unix_time_now < expiration_unix_time:
            return False
        else:
            return True

    def save_dict_to_local(self, df_dict, pickle_path):
        with open(pickle_path, 'wb') as f:
            pickle.dump(df_dict, f)

    def read_dict_from_local(self, pickle_path):
        try:
            with open(pickle_path, 'rb') as handle:
                df_dict = pickle.load(handle)
        except Exception as exc:
            df_dict = None
        return df_dict

